/***************************************************
MODULO: sem.c
SCOPO:  Illustrare il funz. di semop() e semctl()
USO:    Lanciare il programma sem.x e quindi 
        semop.x o semctl.x in una shell separata
****************************************************/

#include    <stdio.h>
#include    <stdlib.h>
#include    <sys/types.h>
#include    <sys/ipc.h>
#include    <sys/sem.h>

#define KEY 14

int main(int argc, char *argv[]) {

  int semid;  /* identificatore dei semafori */
  int i;  
  /* struttura per le operazioni sui semafori */
  struct sembuf * sops = (struct sembuf *) malloc (sizeof(struct sembuf));
  
  /* 
    Creazione di due semafori con permessi di lettura e scrittura per tutti.
    Le flag ICP_CREAT e IPC_EXCL fanno si che la funzione semget ritorni
    errore se esiste gi un vettore di semafori con chiave KEY.
    Vedere man semget.
  */
  if((semid = semget(KEY, 2, IPC_CREAT | IPC_EXCL | 0666)) == -1) {
    perror("semget");
    exit(1);
  }

  /* Inizializzo i due semafori a 1 */
    
  /* 
    Per eseguire operazioni ATOMICHE sui semafori si usa la funzione
    semop(). Vedere man semop.
    Alla funzione semop() vengono passati 3 argomenti:
    - l'identificatore dell'array di semafori su cui eseguire l'operazione
    - il puntatore alla struttura sembuf necessaria per eseguire le operazioni
    - il numero di operazioni da eseguire
      
    Per ogni operazione da eseguire  necessario creare una struttura di tipo
    sembuf. La struttura contiene 3 campi:
    - il numero del semaforo da utilizzare. Ricordare che la semget ritorna
      array di semafori.
    - un intero N che rappresenta l'operazione da eseguire.
      Se l'intero N e' > 0 il valore del semaforo viene incrementato di tale
      quantit. Se N = 0 la semop blocca il processo in attesa che il valore
      del semaforo diventi 0. Se N < 0 la semop blocca il processo in attesa
      che il valore del semaforo meno N sia maggiore o uguale a 0.
    - una eventuale flag (IPC_NOWAIT o SEM_UNDO)
      IPC_NOWAIT serve per avere semafori non bloccanti
      SEM_UNDO serve per ripristinare il vecchio valore del semaforo in caso
               di terminazioni precoci.
  */
      
    sops[0].sem_num = 0;  /* semaforo 0 */
    sops[0].sem_op = 1;   /* incrementa di uno il valore del semaforo 0 */
    sops[0].sem_flg = 0;  /* nessuna flag settata */
    /* esecuzione dell'operazione sul semaforo 0 */ 
    semop(semid, sops, 1);  
  
    sops[0].sem_num = 1;  /* semaforo 1 */
    sops[0].sem_op = 1;   /* incrementa di uno il valore del semaforo 1 */
    sops[0].sem_flg = 0;
    /* esecuzione dell'operazione sul semaforo 1 */ 
    semop(semid, sops, 1);  

    printf("I semafori sono stati inizializzati a 1.\n");
    printf("Lanciare il programma semctl.x o semop.x su un'altra shell             e fornire semid=%d\n", semid);   
    
    while(1) {
    
      sops->sem_num = 0;  /* semaforo 0 */
      sops->sem_op = 0;   /* attende che il semaforo valga zero */
      sops->sem_flg = 0;
      /* quando viene eseguita questa operazione il codice 
         si blocca in attesa che il valore del semaforo 0 diventi 0 */
      semop(semid, sops, 1);
      /* Quando il semaforo diventa 0 stampo che il processo
         e' stato sbloccato */
      printf("Sbloccato 1\n");
      
      sops->sem_num = 1;  /* semaforo 1 */ 
      sops->sem_op = 0;   /* attende che il semaforo valga zero */
      sops->sem_flg = 0; /* notare il comportamento diverso del semaforo 1*/
    
      /* quando viene eseguita questa operazione il codice 
       si blocca in attesa che il valore del semaforo 1 diventi 0 */
      semop(semid, sops, 1); 
      /* Quando il semaforo diventa 0 stampo che il processo
         e' stato sbloccato */      
      printf("              Sbloccato 2\n");
      
    }
}
