/*************************************************
MODULO: shmctl.c
SCOPO: Illustrare il funz. di shmctl()
USO: Lanciare il programma e fornire l'ID
     di un segmento di memoria condivisa
     precedentemente creato.
     Usare il comando della shell ipcs per vedere
     i segmenti di memoria condivisa attivi
**************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <time.h>

void do_shmctl();

int main(int argc, char *argv[]) {
  int cmd; /* comando per shmctl() */
  int shmid; /* ID dell'area di memoria condivisa*/
  struct shmid_ds shmid_ds; /* struttura per il controllo 
                               dell'area di memoria condivisa*/

  fprintf(stderr, "Inserire l'ID del segmento di memoria condiviso: ");
  scanf("%i", &shmid);

  fprintf(stderr, "Comandi validi:\n");
  fprintf(stderr, "\tIPC_RMID =\t%d\n", IPC_RMID);
  fprintf(stderr, "\tIPC_SET =\t%d\n", IPC_SET);
  fprintf(stderr, "\tIPC_STAT =\t%d\n", IPC_STAT);
  fprintf(stderr, "\tSHM_LOCK =\t%d\n", SHM_LOCK);
  fprintf(stderr, "\tSHM_UNLOCK =\t%d\n", SHM_UNLOCK);
  fprintf(stderr, "Scegliere il comando desiderato: ");
  scanf("%i", &cmd);

  switch (cmd) {
    case IPC_STAT:
      /* Le informazioni sullo stato della memoria condivisa
      vengono recuperate con la chiamata alla funzione
      do_shmctl(shmid, cmd, &shmid_ds) eseguita al termine
      del case. Le informazioni saranno inserite nella
      struttura shmid_ds*/
      break;
    case IPC_SET:
      /*visualizzazione dello stato attuale della memoria */
      do_shmctl(shmid, IPC_STAT, &shmid_ds);
      /* Lettura da tastiera dei valori di UID, GID, e permessi da settare */
      fprintf(stderr, "\nInserire shm_perm.uid: ");
      scanf("%hi", &shmid_ds.shm_perm.uid);
      fprintf(stderr, "Inserire shm_perm.gid: ");
      scanf("%hi", &shmid_ds.shm_perm.gid);
      fprintf(stderr, "N.B.: Mantieni il permesso di lettura per te stesso!\n");
      fprintf(stderr, "Inserire shm_perm.mode: ");
      scanf("%hi", &shmid_ds.shm_perm.mode);
      break;
    case IPC_RMID: /* Rimuove il segmento */
      break;
    case SHM_LOCK: /* Esegui il lock sul segmento */
      break;
    case SHM_UNLOCK: /* Esegui unlock sul segmento */
      break;
    default: /* Comando sconosciuto passato a do_shmctl */
      break;
  }
  /* La funzione do_shmctl esegue il comando scelto dall'utente */
  do_shmctl(shmid, cmd, &shmid_ds);
  exit(0);
}

void do_shmctl(int shmid, int cmd, struct shmid_ds* buf) {
  int rtrn; /* valore di ritorno della shmctl */

  fprintf(stderr, "shmctl: Chiamo shmctl(%d, %d, buf)\n", shmid, cmd);
  if (cmd == IPC_SET) {
    fprintf(stderr, "\tbuf->shm_perm.uid == %d\n", buf->shm_perm.uid);
    fprintf(stderr, "\tbuf->shm_perm.gid == %d\n", buf->shm_perm.gid);
    fprintf(stderr, "\tbuf->shm_perm.mode == %#o\n", buf->shm_perm.mode);
  }
  if ((rtrn = shmctl(shmid, cmd, buf)) == -1) {
    perror("shmctl: shmctl fallita.");
    exit(1);
  } else {
    fprintf(stderr, "shmctl: shmctl ha ritornato %d\n", rtrn);
  }
  if (cmd != IPC_STAT && cmd != IPC_SET)
    return; /* ritorno perche' il comando e' stato eseguito e non
               devo visualizzare nessuna informazione sullo stato */

  /* Stampa lo stato corrente del segmento */
  fprintf(stderr, "\nCurrent status:\n");
  fprintf(stderr, "\tshm_perm.uid = %d\n", buf->shm_perm.uid);
  fprintf(stderr, "\tshm_perm.gid = %d\n", buf->shm_perm.gid);
  fprintf(stderr, "\tshm_perm.cuid = %d\n", buf->shm_perm.cuid);
  fprintf(stderr, "\tshm_perm.cgid = %d\n", buf->shm_perm.cgid);
  fprintf(stderr, "\tshm_perm.mode = %#o\n", buf->shm_perm.mode);
  fprintf(stderr, "\tshm_perm.key = %#x\n", buf->shm_perm.__key);
  fprintf(stderr, "\tshm_segsz = %d\n", buf->shm_segsz);
  fprintf(stderr, "\tshm_lpid = %d\n", buf->shm_lpid);
  fprintf(stderr, "\tshm_cpid = %d\n", buf->shm_cpid);
  fprintf(stderr, "\tshm_nattch = %d\n", buf->shm_nattch);
  if (buf->shm_atime) 
    fprintf(stderr, "\tshm_atime = %s", ctime(&buf->shm_atime));
  if (buf->shm_dtime)
    fprintf(stderr, "\tshm_dtime = %s", ctime(&buf->shm_dtime));
  fprintf(stderr, "\tshm_ctime = %s", ctime(&buf->shm_ctime));
}
